@@ -4,6 +4,9 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
+#include "nsURLHelper.h"
+
+#include "mozilla/Encoding.h"
 #include "mozilla/RangedPtr.h"
 #include "mozilla/TextUtils.h"
 
@@ -11,7 +14,6 @@
 #include <iterator>
 
 #include "nsASCIIMask.h"
-#include "nsURLHelper.h"
 #include "nsIFile.h"
 #include "nsIURLParser.h"
 #include "nsCOMPtr.h"
@@ -19,9 +21,12 @@
 #include "nsNetCID.h"
 #include "mozilla/Preferences.h"
 #include "prnetdb.h"
+#include "mozilla/StaticPrefs_network.h"
 #include "mozilla/Tokenizer.h"
 #include "nsEscape.h"
-#include "rust-helper/src/helper.h"
+#include "nsDOMString.h"
+#include "mozilla/net/rust_helper.h"
+#include "mozilla/net/DNS.h"
 
 using namespace mozilla;
 
@@ -30,10 +35,9 @@ using namespace mozilla;
 //----------------------------------------------------------------------------
 
 static bool gInitialized = false;
-static nsIURLParser* gNoAuthURLParser = nullptr;
-static nsIURLParser* gAuthURLParser = nullptr;
-static nsIURLParser* gStdURLParser = nullptr;
-static int32_t gMaxLength = 1048576;  // Default: 1MB
+static StaticRefPtr<nsIURLParser> gNoAuthURLParser;
+static StaticRefPtr<nsIURLParser> gAuthURLParser;
+static StaticRefPtr<nsIURLParser> gStdURLParser;
 
 static void InitGlobals() {
   nsCOMPtr<nsIURLParser> parser;
@@ -41,40 +45,33 @@ static void InitGlobals() {
   parser = do_GetService(NS_NOAUTHURLPARSER_CONTRACTID);
   NS_ASSERTION(parser, "failed getting 'noauth' url parser");
   if (parser) {
-    gNoAuthURLParser = parser.get();
-    NS_ADDREF(gNoAuthURLParser);
+    gNoAuthURLParser = parser;
   }
 
   parser = do_GetService(NS_AUTHURLPARSER_CONTRACTID);
   NS_ASSERTION(parser, "failed getting 'auth' url parser");
   if (parser) {
-    gAuthURLParser = parser.get();
-    NS_ADDREF(gAuthURLParser);
+    gAuthURLParser = parser;
   }
 
   parser = do_GetService(NS_STDURLPARSER_CONTRACTID);
   NS_ASSERTION(parser, "failed getting 'std' url parser");
   if (parser) {
-    gStdURLParser = parser.get();
-    NS_ADDREF(gStdURLParser);
+    gStdURLParser = parser;
   }
 
   gInitialized = true;
-  Preferences::AddIntVarCache(&gMaxLength, "network.standard-url.max-length",
-                              1048576);
 }
 
 void net_ShutdownURLHelper() {
   if (gInitialized) {
-    NS_IF_RELEASE(gNoAuthURLParser);
-    NS_IF_RELEASE(gAuthURLParser);
-    NS_IF_RELEASE(gStdURLParser);
     gInitialized = false;
   }
+  gNoAuthURLParser = nullptr;
+  gAuthURLParser = nullptr;
+  gStdURLParser = nullptr;
 }
 
-int32_t net_GetURLMaxLength() { return gMaxLength; }
-
 //----------------------------------------------------------------------------
 // nsIURLParser getters
 //----------------------------------------------------------------------------
@@ -139,7 +136,8 @@ nsresult net_ParseFileURL(const nsACString& inURL, nsACString& outDirectory,
                           nsACString& outFileExtension) {
   nsresult rv;
 
-  if (inURL.Length() > (uint32_t)gMaxLength) {
+  if (inURL.Length() >
+      (uint32_t)StaticPrefs::network_standard_url_max_length()) {
     return NS_ERROR_MALFORMED_URI;
   }
 
@@ -186,13 +184,16 @@ nsresult net_ParseFileURL(const nsACString& inURL, nsACString& outDirectory,
                              &extensionPos, &extensionLen);
   if (NS_FAILED(rv)) return rv;
 
-  if (directoryLen > 0)
+  if (directoryLen > 0) {
     outDirectory = Substring(inURL, filepathPos + directoryPos, directoryLen);
-  if (basenameLen > 0)
+  }
+  if (basenameLen > 0) {
     outFileBaseName = Substring(inURL, filepathPos + basenamePos, basenameLen);
-  if (extensionLen > 0)
+  }
+  if (extensionLen > 0) {
     outFileExtension =
         Substring(inURL, filepathPos + extensionPos, extensionLen);
+  }
   // since we are using a no-auth url parser, there will never be a host
   // XXX not strictly true... file://localhost/foo/bar.html is a valid URL
 
@@ -218,16 +219,22 @@ void net_CoalesceDirs(netCoalesceFlags flags, char* path) {
   uint32_t traversal = 0;
   uint32_t special_ftp_len = 0;
 
+  MOZ_ASSERT(*path == '/', "We expect the path to begin with /");
+  if (*path != '/') {
+    return;
+  }
+
   /* Remember if this url is a special ftp one: */
   if (flags & NET_COALESCE_DOUBLE_SLASH_IS_ROOT) {
     /* some schemes (for example ftp) have the speciality that
        the path can begin // or /%2F to mark the root of the
        servers filesystem, a simple / only marks the root relative
        to the user loging in. We remember the length of the marker */
-    if (nsCRT::strncasecmp(path, "/%2F", 4) == 0)
+    if (nsCRT::strncasecmp(path, "/%2F", 4) == 0) {
       special_ftp_len = 4;
-    else if (strncmp(path, "//", 2) == 0)
+    } else if (strncmp(path, "//", 2) == 0) {
       special_ftp_len = 2;
+    }
   }
 
   /* find the last slash before # or ? */
@@ -247,7 +254,7 @@ void net_CoalesceDirs(netCoalesceFlags flags, char* path) {
   fwdPtr = path;
 
   /* replace all %2E or %2e with . in the path */
-  /* but stop at lastchar if non null */
+  /* but stop at lastslash if non null */
   for (; (*fwdPtr != '\0') && (*fwdPtr != '?') && (*fwdPtr != '#') &&
          (*lastslash == '\0' || fwdPtr != lastslash);
        ++fwdPtr) {
@@ -285,8 +292,9 @@ void net_CoalesceDirs(netCoalesceFlags flags, char* path) {
       // otherwise retain them in the path
       if (traversal > 0 || !(flags & NET_COALESCE_ALLOW_RELATIVE_ROOT)) {
         if (urlPtr != path) urlPtr--;  // we must be going back at least by one
-        for (; *urlPtr != '/' && urlPtr != path; urlPtr--)
-          ;           // null body
+        for (; *urlPtr != '/' && urlPtr != path; urlPtr--) {
+          ;  // null body
+        }
         --traversal;  // count back
         // forward the fwdPtr past the ../
         fwdPtr += 2;
@@ -311,10 +319,11 @@ void net_CoalesceDirs(netCoalesceFlags flags, char* path) {
         // /%2F and urlPtr just points at the "F" of "/%2F" then do
         // not overwrite it with the /, just copy .. and move forward
         // urlPtr.
-        if (special_ftp_len > 3 && urlPtr == path + special_ftp_len - 1)
+        if (special_ftp_len > 3 && urlPtr == path + special_ftp_len - 1) {
           ++urlPtr;
-        else
+        } else {
           *urlPtr++ = *fwdPtr;
+        }
         ++fwdPtr;
         *urlPtr++ = *fwdPtr;
         ++fwdPtr;
@@ -324,8 +333,9 @@ void net_CoalesceDirs(netCoalesceFlags flags, char* path) {
       // count the hierachie, but only if we do not have reached
       // the root of some special urls with a special root marker
       if (*fwdPtr == '/' && *(fwdPtr + 1) != '.' &&
-          (special_ftp_len != 2 || *(fwdPtr + 1) != '/'))
+          (special_ftp_len != 2 || *(fwdPtr + 1) != '/')) {
         traversal++;
+      }
       // copy the url incrementaly
       *urlPtr++ = *fwdPtr;
     }
@@ -336,8 +346,18 @@ void net_CoalesceDirs(netCoalesceFlags flags, char* path) {
    *     /foo/foo1/.   ->  /foo/foo1/
    */
 
-  if ((urlPtr > (path + 1)) && (*(urlPtr - 1) == '.') && (*(urlPtr - 2) == '/'))
+  if ((urlPtr > (path + 1)) && (*(urlPtr - 1) == '.') &&
+      (*(urlPtr - 2) == '/')) {
     urlPtr--;
+  }
+
+  // Before we start copying past ?#, we must make sure we don't overwrite
+  // the first / character.  If fwdPtr is also unchanged, just copy everything
+  // (this shouldn't happen unless we could get in here without a leading
+  // slash).
+  if (urlPtr == path && fwdPtr != path) {
+    urlPtr++;
+  }
 
   // Copy remaining stuff past the #?;
   for (; *fwdPtr != '\0'; ++fwdPtr) {
@@ -346,80 +366,12 @@ void net_CoalesceDirs(netCoalesceFlags flags, char* path) {
   *urlPtr = '\0';  // terminate the url
 }
 
-nsresult net_ResolveRelativePath(const nsACString& relativePath,
-                                 const nsACString& basePath,
-                                 nsACString& result) {
-  nsAutoCString name;
-  nsAutoCString path(basePath);
-  bool needsDelim = false;
-
-  if (!path.IsEmpty()) {
-    char16_t last = path.Last();
-    needsDelim = !(last == '/');
-  }
-
-  nsACString::const_iterator beg, end;
-  relativePath.BeginReading(beg);
-  relativePath.EndReading(end);
-
-  bool stop = false;
-  char c;
-  for (; !stop; ++beg) {
-    c = (beg == end) ? '\0' : *beg;
-    // printf("%c [name=%s] [path=%s]\n", c, name.get(), path.get());
-    switch (c) {
-      case '\0':
-      case '#':
-      case '?':
-        stop = true;
-        MOZ_FALLTHROUGH;
-      case '/':
-        // delimiter found
-        if (name.EqualsLiteral("..")) {
-          // pop path
-          // If we already have the delim at end, then
-          //  skip over that when searching for next one to the left
-          int32_t offset = path.Length() - (needsDelim ? 1 : 2);
-          // First check for errors
-          if (offset < 0) return NS_ERROR_MALFORMED_URI;
-          int32_t pos = path.RFind("/", false, offset);
-          if (pos >= 0)
-            path.Truncate(pos + 1);
-          else
-            path.Truncate();
-        } else if (name.IsEmpty() || name.EqualsLiteral(".")) {
-          // do nothing
-        } else {
-          // append name to path
-          if (needsDelim) path += '/';
-          path += name;
-          needsDelim = true;
-        }
-        name.Truncate();
-        break;
-
-      default:
-        // append char to name
-        name += c;
-    }
-  }
-  // append anything left on relativePath (e.g. #..., ;..., ?...)
-  if (c != '\0') path += Substring(--beg, end);
-
-  result = path;
-  return NS_OK;
-}
-
 //----------------------------------------------------------------------------
 // scheme fu
 //----------------------------------------------------------------------------
 
 static bool net_IsValidSchemeChar(const char aChar) {
-  if (IsAsciiAlpha(aChar) || IsAsciiDigit(aChar) || aChar == '+' ||
-      aChar == '.' || aChar == '-') {
-    return true;
-  }
-  return false;
+  return mozilla::net::rust_net_is_valid_scheme_char(aChar);
 }
 
 /* Extract URI-Scheme if possible */
@@ -457,18 +409,8 @@ nsresult net_ExtractURLScheme(const nsACString& inURI, nsACString& scheme) {
   return NS_OK;
 }
 
-bool net_IsValidScheme(const char* scheme, uint32_t schemeLen) {
-  // first char must be alpha
-  if (!IsAsciiAlpha(*scheme)) return false;
-
-  // nsCStrings may have embedded nulls -- reject those too
-  for (; schemeLen; ++scheme, --schemeLen) {
-    if (!(IsAsciiAlpha(*scheme) || IsAsciiDigit(*scheme) || *scheme == '+' ||
-          *scheme == '.' || *scheme == '-'))
-      return false;
-  }
-
-  return true;
+bool net_IsValidScheme(const nsACString& scheme) {
+  return mozilla::net::rust_net_is_valid_scheme(&scheme);
 }
 
 bool net_IsAbsoluteURL(const nsACString& uri) {
@@ -514,13 +456,13 @@ bool net_IsAbsoluteURL(const nsACString& uri) {
 void net_FilterURIString(const nsACString& input, nsACString& result) {
   result.Truncate();
 
-  auto start = input.BeginReading();
-  auto end = input.EndReading();
+  const auto* start = input.BeginReading();
+  const auto* end = input.EndReading();
 
   // Trim off leading and trailing invalid chars.
   auto charFilter = [](char c) { return static_cast<uint8_t>(c) > 0x20; };
-  auto newStart = std::find_if(start, end, charFilter);
-  auto newEnd =
+  const auto* newStart = std::find_if(start, end, charFilter);
+  const auto* newEnd =
       std::find_if(std::reverse_iterator<decltype(end)>(end),
                    std::reverse_iterator<decltype(newStart)>(newStart),
                    charFilter)
@@ -529,7 +471,7 @@ void net_FilterURIString(const nsACString& input, nsACString& result) {
   // Check if chars need to be stripped.
   bool needsStrip = false;
   const ASCIIMaskArray& mask = ASCIIMask::MaskCRLFTab();
-  for (auto itr = start; itr != end; ++itr) {
+  for (const auto* itr = start; itr != end; ++itr) {
     if (ASCIIMask::IsMasked(mask, *itr)) {
       needsStrip = true;
       break;
@@ -550,24 +492,24 @@ void net_FilterURIString(const nsACString& input, nsACString& result) {
 }
 
 nsresult net_FilterAndEscapeURI(const nsACString& aInput, uint32_t aFlags,
+                                const ASCIIMaskArray& aFilterMask,
                                 nsACString& aResult) {
   aResult.Truncate();
 
-  auto start = aInput.BeginReading();
-  auto end = aInput.EndReading();
+  const auto* start = aInput.BeginReading();
+  const auto* end = aInput.EndReading();
 
   // Trim off leading and trailing invalid chars.
   auto charFilter = [](char c) { return static_cast<uint8_t>(c) > 0x20; };
-  auto newStart = std::find_if(start, end, charFilter);
-  auto newEnd =
+  const auto* newStart = std::find_if(start, end, charFilter);
+  const auto* newEnd =
       std::find_if(std::reverse_iterator<decltype(end)>(end),
                    std::reverse_iterator<decltype(newStart)>(newStart),
                    charFilter)
           .base();
 
-  const ASCIIMaskArray& mask = ASCIIMask::MaskCRLFTab();
-  return NS_EscapeAndFilterURL(Substring(newStart, newEnd), aFlags, &mask,
-                               aResult, fallible);
+  return NS_EscapeAndFilterURL(Substring(newStart, newEnd), aFlags,
+                               &aFilterMask, aResult, fallible);
 }
 
 #if defined(XP_WIN)
@@ -587,6 +529,11 @@ bool net_NormalizeFileURL(const nsACString& aURL, nsCString& aResultBuf) {
       aResultBuf += '/';
       begin = s + 1;
     }
+    if (*s == '#') {
+      // Don't normalize any backslashes following the hash.
+      s = endIter.get();
+      break;
+    }
   }
   if (writing && s > begin) aResultBuf.Append(begin, s - begin);
 
@@ -727,11 +674,9 @@ static void net_ParseMediaType(const nsACString& aMediaTypeStr,
   const char* start = flatStr.get();
   const char* end = start + flatStr.Length();
 
-  // Trim LWS leading and trailing whitespace from type.  We include '(' in
-  // the trailing trim set to catch media-type comments, which are not at all
-  // standard, but may occur in rare cases.
+  // Trim LWS leading and trailing whitespace from type.
   const char* type = net_FindCharNotInSet(start, end, HTTP_LWS);
-  const char* typeEnd = net_FindCharInSet(type, end, HTTP_LWS ";(");
+  const char* typeEnd = net_FindCharInSet(type, end, HTTP_LWS ";");
 
   const char* charset = "";
   const char* charsetEnd = charset;
@@ -753,7 +698,8 @@ static void net_ParseMediaType(const nsACString& aMediaTypeStr,
       const char* paramName = net_FindCharNotInSet(
           start + curParamStart, start + curParamEnd, HTTP_LWS);
       static const char charsetStr[] = "charset=";
-      if (PL_strncasecmp(paramName, charsetStr, sizeof(charsetStr) - 1) == 0) {
+      if (nsCRT::strncasecmp(paramName, charsetStr, sizeof(charsetStr) - 1) ==
+          0) {
         charset = paramName + sizeof(charsetStr) - 1;
         charsetEnd = start + curParamEnd;
         typeHasCharset = true;
@@ -768,9 +714,7 @@ static void net_ParseMediaType(const nsACString& aMediaTypeStr,
 
   bool charsetNeedsQuotedStringUnescaping = false;
   if (typeHasCharset) {
-    // Trim LWS leading and trailing whitespace from charset.  We include
-    // '(' in the trailing trim set to catch media-type comments, which are
-    // not at all standard, but may occur in rare cases.
+    // Trim LWS leading and trailing whitespace from charset.
     charset = net_FindCharNotInSet(charset, charsetEnd, HTTP_LWS);
     if (*charset == '"') {
       charsetNeedsQuotedStringUnescaping = true;
@@ -779,7 +723,7 @@ static void net_ParseMediaType(const nsACString& aMediaTypeStr,
       charset++;
       NS_ASSERTION(charsetEnd >= charset, "Bad charset parsing");
     } else {
-      charsetEnd = net_FindCharInSet(charset, charsetEnd, HTTP_LWS ";(");
+      charsetEnd = net_FindCharInSet(charset, charsetEnd, HTTP_LWS ";");
     }
   }
 
@@ -797,7 +741,7 @@ static void net_ParseMediaType(const nsACString& aMediaTypeStr,
     // Common case here is that aContentType is empty
     bool eq = !aContentType.IsEmpty() &&
               aContentType.Equals(Substring(type, typeEnd),
-                                  nsCaseInsensitiveCStringComparator());
+                                  nsCaseInsensitiveCStringComparator);
     if (!eq) {
       aContentType.Assign(type, typeEnd - type);
       ToLowerCase(aContentType);
@@ -942,6 +886,11 @@ void net_ParseRequestContentType(const nsACString& aHeaderStr,
 }
 
 bool net_IsValidHostName(const nsACString& host) {
+  // The host name is limited to 253 ascii characters.
+  if (host.Length() > 253) {
+    return false;
+  }
+
   const char* end = host.EndReading();
   // Use explicit whitelists to select which characters we are
   // willing to send to lower-level DNS logic. This is more
@@ -956,19 +905,465 @@ bool net_IsValidHostName(const nsACString& host) {
   if (net_FindCharNotInSet(host.BeginReading(), end,
                            "abcdefghijklmnopqrstuvwxyz"
                            ".-0123456789"
-                           "ABCDEFGHIJKLMNOPQRSTUVWXYZ$+_") == end)
+                           "ABCDEFGHIJKLMNOPQRSTUVWXYZ$+_") == end) {
     return true;
+  }
 
   // Might be a valid IPv6 link-local address containing a percent sign
-  nsAutoCString strhost(host);
-  PRNetAddr addr;
-  return PR_StringToNetAddr(strhost.get(), &addr) == PR_SUCCESS;
+  return mozilla::net::HostIsIPLiteral(host);
 }
 
 bool net_IsValidIPv4Addr(const nsACString& aAddr) {
-  return rust_net_is_valid_ipv4_addr(aAddr);
+  return mozilla::net::rust_net_is_valid_ipv4_addr(&aAddr);
 }
 
 bool net_IsValidIPv6Addr(const nsACString& aAddr) {
-  return rust_net_is_valid_ipv6_addr(aAddr);
+  return mozilla::net::rust_net_is_valid_ipv6_addr(&aAddr);
 }
+
+bool net_GetDefaultStatusTextForCode(uint16_t aCode, nsACString& aOutText) {
+  switch (aCode) {
+      // start with the most common
+    case 200:
+      aOutText.AssignLiteral("OK");
+      break;
+    case 404:
+      aOutText.AssignLiteral("Not Found");
+      break;
+    case 301:
+      aOutText.AssignLiteral("Moved Permanently");
+      break;
+    case 304:
+      aOutText.AssignLiteral("Not Modified");
+      break;
+    case 307:
+      aOutText.AssignLiteral("Temporary Redirect");
+      break;
+    case 500:
+      aOutText.AssignLiteral("Internal Server Error");
+      break;
+
+      // also well known
+    case 100:
+      aOutText.AssignLiteral("Continue");
+      break;
+    case 101:
+      aOutText.AssignLiteral("Switching Protocols");
+      break;
+    case 201:
+      aOutText.AssignLiteral("Created");
+      break;
+    case 202:
+      aOutText.AssignLiteral("Accepted");
+      break;
+    case 203:
+      aOutText.AssignLiteral("Non Authoritative");
+      break;
+    case 204:
+      aOutText.AssignLiteral("No Content");
+      break;
+    case 205:
+      aOutText.AssignLiteral("Reset Content");
+      break;
+    case 206:
+      aOutText.AssignLiteral("Partial Content");
+      break;
+    case 207:
+      aOutText.AssignLiteral("Multi-Status");
+      break;
+    case 208:
+      aOutText.AssignLiteral("Already Reported");
+      break;
+    case 300:
+      aOutText.AssignLiteral("Multiple Choices");
+      break;
+    case 302:
+      aOutText.AssignLiteral("Found");
+      break;
+    case 303:
+      aOutText.AssignLiteral("See Other");
+      break;
+    case 305:
+      aOutText.AssignLiteral("Use Proxy");
+      break;
+    case 308:
+      aOutText.AssignLiteral("Permanent Redirect");
+      break;
+    case 400:
+      aOutText.AssignLiteral("Bad Request");
+      break;
+    case 401:
+      aOutText.AssignLiteral("Unauthorized");
+      break;
+    case 402:
+      aOutText.AssignLiteral("Payment Required");
+      break;
+    case 403:
+      aOutText.AssignLiteral("Forbidden");
+      break;
+    case 405:
+      aOutText.AssignLiteral("Method Not Allowed");
+      break;
+    case 406:
+      aOutText.AssignLiteral("Not Acceptable");
+      break;
+    case 407:
+      aOutText.AssignLiteral("Proxy Authentication Required");
+      break;
+    case 408:
+      aOutText.AssignLiteral("Request Timeout");
+      break;
+    case 409:
+      aOutText.AssignLiteral("Conflict");
+      break;
+    case 410:
+      aOutText.AssignLiteral("Gone");
+      break;
+    case 411:
+      aOutText.AssignLiteral("Length Required");
+      break;
+    case 412:
+      aOutText.AssignLiteral("Precondition Failed");
+      break;
+    case 413:
+      aOutText.AssignLiteral("Request Entity Too Large");
+      break;
+    case 414:
+      aOutText.AssignLiteral("Request URI Too Long");
+      break;
+    case 415:
+      aOutText.AssignLiteral("Unsupported Media Type");
+      break;
+    case 416:
+      aOutText.AssignLiteral("Requested Range Not Satisfiable");
+      break;
+    case 417:
+      aOutText.AssignLiteral("Expectation Failed");
+      break;
+    case 418:
+      aOutText.AssignLiteral("I'm a teapot");
+      break;
+    case 421:
+      aOutText.AssignLiteral("Misdirected Request");
+      break;
+    case 422:
+      aOutText.AssignLiteral("Unprocessable Entity");
+      break;
+    case 423:
+      aOutText.AssignLiteral("Locked");
+      break;
+    case 424:
+      aOutText.AssignLiteral("Failed Dependency");
+      break;
+    case 425:
+      aOutText.AssignLiteral("Too Early");
+      break;
+    case 426:
+      aOutText.AssignLiteral("Upgrade Required");
+      break;
+    case 428:
+      aOutText.AssignLiteral("Precondition Required");
+      break;
+    case 429:
+      aOutText.AssignLiteral("Too Many Requests");
+      break;
+    case 431:
+      aOutText.AssignLiteral("Request Header Fields Too Large");
+      break;
+    case 451:
+      aOutText.AssignLiteral("Unavailable For Legal Reasons");
+      break;
+    case 501:
+      aOutText.AssignLiteral("Not Implemented");
+      break;
+    case 502:
+      aOutText.AssignLiteral("Bad Gateway");
+      break;
+    case 503:
+      aOutText.AssignLiteral("Service Unavailable");
+      break;
+    case 504:
+      aOutText.AssignLiteral("Gateway Timeout");
+      break;
+    case 505:
+      aOutText.AssignLiteral("HTTP Version Unsupported");
+      break;
+    case 506:
+      aOutText.AssignLiteral("Variant Also Negotiates");
+      break;
+    case 507:
+      aOutText.AssignLiteral("Insufficient Storage ");
+      break;
+    case 508:
+      aOutText.AssignLiteral("Loop Detected");
+      break;
+    case 510:
+      aOutText.AssignLiteral("Not Extended");
+      break;
+    case 511:
+      aOutText.AssignLiteral("Network Authentication Required");
+      break;
+    default:
+      aOutText.AssignLiteral("No Reason Phrase");
+      return false;
+  }
+  return true;
+}
+
+static auto MakeNameMatcher(const nsACString& aName) {
+  return [&aName](const auto& param) { return param.mKey.Equals(aName); };
+}
+
+static void AssignMaybeInvalidUTF8String(const nsACString& aSource,
+                                         nsACString& aDest) {
+  if (NS_FAILED(UTF_8_ENCODING->DecodeWithoutBOMHandling(aSource, aDest))) {
+    MOZ_CRASH("Out of memory when converting URL params.");
+  }
+}
+
+namespace mozilla {
+
+bool URLParams::Has(const nsACString& aName) {
+  return std::any_of(mParams.cbegin(), mParams.cend(), MakeNameMatcher(aName));
+}
+
+bool URLParams::Has(const nsACString& aName, const nsACString& aValue) {
+  return std::any_of(
+      mParams.cbegin(), mParams.cend(), [&aName, &aValue](const auto& param) {
+        return param.mKey.Equals(aName) && param.mValue.Equals(aValue);
+      });
+}
+
+void URLParams::Get(const nsACString& aName, nsACString& aRetval) {
+  aRetval.SetIsVoid(true);
+
+  const auto end = mParams.cend();
+  const auto it = std::find_if(mParams.cbegin(), end, MakeNameMatcher(aName));
+  if (it != end) {
+    aRetval.Assign(it->mValue);
+  }
+}
+
+void URLParams::GetAll(const nsACString& aName, nsTArray<nsCString>& aRetval) {
+  aRetval.Clear();
+
+  for (uint32_t i = 0, len = mParams.Length(); i < len; ++i) {
+    if (mParams[i].mKey.Equals(aName)) {
+      aRetval.AppendElement(mParams[i].mValue);
+    }
+  }
+}
+
+void URLParams::Append(const nsACString& aName, const nsACString& aValue) {
+  Param* param = mParams.AppendElement();
+  param->mKey = aName;
+  param->mValue = aValue;
+}
+
+void URLParams::Set(const nsACString& aName, const nsACString& aValue) {
+  Param* param = nullptr;
+  for (uint32_t i = 0, len = mParams.Length(); i < len;) {
+    if (!mParams[i].mKey.Equals(aName)) {
+      ++i;
+      continue;
+    }
+    if (!param) {
+      param = &mParams[i];
+      ++i;
+      continue;
+    }
+    // Remove duplicates.
+    mParams.RemoveElementAt(i);
+    --len;
+  }
+
+  if (!param) {
+    param = mParams.AppendElement();
+    param->mKey = aName;
+  }
+
+  param->mValue = aValue;
+}
+
+void URLParams::Delete(const nsACString& aName) {
+  mParams.RemoveElementsBy(
+      [&aName](const auto& param) { return param.mKey.Equals(aName); });
+}
+
+void URLParams::Delete(const nsACString& aName, const nsACString& aValue) {
+  mParams.RemoveElementsBy([&aName, &aValue](const auto& param) {
+    return param.mKey.Equals(aName) && param.mValue.Equals(aValue);
+  });
+}
+
+/* static */
+void URLParams::DecodeString(const nsACString& aInput, nsACString& aOutput) {
+  const char* const end = aInput.EndReading();
+  for (const char* iter = aInput.BeginReading(); iter != end;) {
+    // replace '+' with U+0020
+    if (*iter == '+') {
+      aOutput.Append(' ');
+      ++iter;
+      continue;
+    }
+
+    // Percent decode algorithm
+    if (*iter == '%') {
+      const char* const first = iter + 1;
+      const char* const second = first + 1;
+
+      const auto asciiHexDigit = [](char x) {
+        return (x >= 0x41 && x <= 0x46) || (x >= 0x61 && x <= 0x66) ||
+               (x >= 0x30 && x <= 0x39);
+      };
+
+      const auto hexDigit = [](char x) {
+        return x >= 0x30 && x <= 0x39
+                   ? x - 0x30
+                   : (x >= 0x41 && x <= 0x46 ? x - 0x37 : x - 0x57);
+      };
+
+      if (first != end && second != end && asciiHexDigit(*first) &&
+          asciiHexDigit(*second)) {
+        aOutput.Append(hexDigit(*first) * 16 + hexDigit(*second));
+        iter = second + 1;
+      } else {
+        aOutput.Append('%');
+        ++iter;
+      }
+
+      continue;
+    }
+
+    aOutput.Append(*iter);
+    ++iter;
+  }
+  AssignMaybeInvalidUTF8String(aOutput, aOutput);
+}
+
+/* static */
+bool URLParams::ParseNextInternal(const char*& aStart, const char* const aEnd,
+                                  bool aShouldDecode, nsACString* aOutputName,
+                                  nsACString* aOutputValue) {
+  nsDependentCSubstring string;
+
+  const char* const iter = std::find(aStart, aEnd, '&');
+  if (iter != aEnd) {
+    string.Rebind(aStart, iter);
+    aStart = iter + 1;
+  } else {
+    string.Rebind(aStart, aEnd);
+    aStart = aEnd;
+  }
+
+  if (string.IsEmpty()) {
+    return false;
+  }
+
+  const auto* const eqStart = string.BeginReading();
+  const auto* const eqEnd = string.EndReading();
+  const auto* const eqIter = std::find(eqStart, eqEnd, '=');
+
+  nsDependentCSubstring name;
+  nsDependentCSubstring value;
+
+  if (eqIter != eqEnd) {
+    name.Rebind(eqStart, eqIter);
+    value.Rebind(eqIter + 1, eqEnd);
+  } else {
+    name.Rebind(string, 0);
+  }
+
+  if (aShouldDecode) {
+    DecodeString(name, *aOutputName);
+    DecodeString(value, *aOutputValue);
+    return true;
+  }
+
+  AssignMaybeInvalidUTF8String(name, *aOutputName);
+  AssignMaybeInvalidUTF8String(value, *aOutputValue);
+  return true;
+}
+
+/* static */
+bool URLParams::Extract(const nsACString& aInput, const nsACString& aName,
+                        nsACString& aValue) {
+  aValue.SetIsVoid(true);
+  return !URLParams::Parse(
+      aInput, true,
+      [&aName, &aValue](const nsACString& name, nsCString&& value) {
+        if (aName == name) {
+          aValue = std::move(value);
+          return false;
+        }
+        return true;
+      });
+}
+
+void URLParams::ParseInput(const nsACString& aInput) {
+  // Remove all the existing data before parsing a new input.
+  DeleteAll();
+
+  URLParams::Parse(aInput, true, [this](nsCString&& name, nsCString&& value) {
+    mParams.AppendElement(Param{std::move(name), std::move(value)});
+    return true;
+  });
+}
+
+void URLParams::SerializeString(const nsACString& aInput, nsACString& aValue) {
+  const unsigned char* p = (const unsigned char*)aInput.BeginReading();
+  const unsigned char* end = p + aInput.Length();
+
+  while (p != end) {
+    // ' ' to '+'
+    if (*p == 0x20) {
+      aValue.Append(0x2B);
+      // Percent Encode algorithm
+    } else if (*p == 0x2A || *p == 0x2D || *p == 0x2E ||
+               (*p >= 0x30 && *p <= 0x39) || (*p >= 0x41 && *p <= 0x5A) ||
+               *p == 0x5F || (*p >= 0x61 && *p <= 0x7A)) {
+      aValue.Append(*p);
+    } else {
+      aValue.AppendPrintf("%%%.2X", *p);
+    }
+
+    ++p;
+  }
+}
+
+void URLParams::Serialize(nsACString& aValue, bool aEncode) const {
+  aValue.Truncate();
+  bool first = true;
+
+  for (uint32_t i = 0, len = mParams.Length(); i < len; ++i) {
+    if (first) {
+      first = false;
+    } else {
+      aValue.Append('&');
+    }
+
+    // XXX Actually, it's not necessary to build a new string object. Generally,
+    // such cases could just convert each codepoint one-by-one.
+    if (aEncode) {
+      SerializeString(mParams[i].mKey, aValue);
+      aValue.Append('=');
+      SerializeString(mParams[i].mValue, aValue);
+    } else {
+      aValue.Append(mParams[i].mKey);
+      aValue.Append('=');
+      aValue.Append(mParams[i].mValue);
+    }
+  }
+}
+
+void URLParams::Sort() {
+  mParams.StableSort([](const Param& lhs, const Param& rhs) {
+    // FIXME(emilio, bug 1888901): The URLSearchParams.sort() spec requires
+    // comparing by utf-16 code points... That's a bit unfortunate, maybe we
+    // can optimize the string conversions here?
+    return Compare(NS_ConvertUTF8toUTF16(lhs.mKey),
+                   NS_ConvertUTF8toUTF16(rhs.mKey));
+  });
+}
+
+}  // namespace mozilla
